<?php
/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */

/**
 * Cron Job that:
 * - reads scheduled task from esb_schedule table;
 * - instantiates object taking class/interface from task; 
 * - converts object into configured output format;
 * - calls endpoint;
 * - in case of error schedules again the task.
 */

namespace FiloBlu\Esb\Cron;

use Magento\Framework\Webapi\ServiceOutputProcessor;

class SendMessage
{
    /**
     * @var \FiloBlu\Esb\Model\ScheduleFactory
     */     
    protected $_scheduleFactory;

    /**
     * @var \FiloBlu\Esb\Helper\Format\FormatFactory
     */        
    protected $_formatFactory;

    /**
     * @var \Psr\Log\LoggerInterface
     */         
    protected $_logger;
    
    /**
     * @var \Magento\Framework\ObjectManagerInterface
     */     
    protected $_objectManager;
    
    /**
     * @var \Magento\Sales\Model\ProductFactory
     */
    protected $_productFactory;
    
    /**
     * @var \Magento\Sales\Model\Order\ItemFactory
     */
    protected $_orderItemFactory;

    /**
     * @var \Magento\Framework\HTTP\ZendClientFactory
     */     
    protected $_httpClientFactory;

    /**
     * @var \Magento\Framework\Webapi\ServiceOutputProcessor
     */     
    protected $_serviceOutputProcessor;

	/**
     * @var \Magento\Framework\App\Config\ScopeConfigInterface
     */  
    protected $_scopeConfig;

	/**
     * @var string
     */  
    protected $_response;    

    /**
     * @param \FiloBlu\Esb\Model\ScheduleFactory $scheduleFactory
     * @param \FiloBlu\Esb\Helper\Format\FormatFactory $formatFactory
     * @param \Magento\Framework\HTTP\ZendClientFactory $httpClientFactory
     * @param \Psr\Log\LoggerInterface $logger
     * @param \Magento\Framework\ObjectManagerInterface $objectManager
     * @param \Magento\Catalog\Model\ProductFactory $productFactory
     * @param \Magento\Sales\Model\Order\ItemFactory $orderItemFactory
     * @param ServiceOutputProcessor $serviceOutputProcessor
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     */
	public function __construct(
            \FiloBlu\Esb\Model\ScheduleFactory $scheduleFactory,
            \FiloBlu\Esb\Helper\Format\FormatFactory $formatFactory,
            \Magento\Framework\HTTP\ZendClientFactory $httpClientFactory,
            \Psr\Log\LoggerInterface $logger,
            \Magento\Framework\ObjectManagerInterface $objectManager,
            \Magento\Catalog\Model\ProductFactory $productFactory,
            \Magento\Sales\Model\Order\ItemFactory $orderItemFactory,
            ServiceOutputProcessor $serviceOutputProcessor,
            \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
	){
            $this->_scheduleFactory=$scheduleFactory;
            $this->_formatFactory=$formatFactory;
            $this->_logger = $logger;
            $this->_objectManager = $objectManager;
            $this->_productFactory = $productFactory;
            $this->_orderItemFactory = $orderItemFactory;
            $this->_httpClientFactory = $httpClientFactory;
            $this->_serviceOutputProcessor = $serviceOutputProcessor;
            $this->_scopeConfig = $scopeConfig;
	}

    public function execute()
    {

        /** Retrieve the pending task to be executed */
        $scheduleModel = $this->_scheduleFactory->create();
        $scheduleCollection = $scheduleModel->getCollection()
                                ->addFieldToFilter("status", "pending")
                                ->addFieldToFilter("executable_after",["lteq" => new \Zend_Db_Expr("NOW()")])
                                ->setOrder("priority", "ASC");

        /** Instantiate format helper (JSON/XML) */
        $formatData = $this->_formatFactory->create();
    
        foreach($scheduleCollection as $schedule){        
            /** Set schedule status to running to prevent concurrent executions.*/
            $schedule->setStatus("running");
            $schedule->setExecutedAt(new \Zend_Db_Expr("NOW()"));
            $schedule->save();
            
            /** Schedule validation.*/
            $validation=$this->validateSchedule($schedule);
            if(!$validation["result"]){
                $schedule->setStatus("skipped");  
                $schedule->setResponse($validation["reason"]);
                $schedule->save();
                continue;
            }


            $message="";

            /** The scheduled task contains a Repository Interface. Use Repository design pattern to load object. */
            if(!is_null($schedule->getObjectRepositoryInterface())){
                $objectInterface=$schedule->getObjectRepositoryInterface();
                $objectRepository=$this->_objectManager->create($objectInterface);
                $outputObject = $objectRepository->get($schedule->getObjectId());
                
                // Add gift wrap product to order
                if($objectInterface == '\Magento\Sales\Api\OrderRepositoryInterface' && $outputObject->getGwId()):
                     $this->addGiftWrapProduct($outputObject);
                endif;

                /* Use serviceOutputProcessor to build output data. See function processApiRequest() of Magento\Webapi\Controller\Rest. */
                $message = $this->_serviceOutputProcessor->process(
                    $outputObject,
                    $objectInterface,
                    "get"
                );                
            }
            /** The scheduled task contains an Object Class. Use standard magento loading. */
            elseif(!is_null($schedule->getObjectClass())){
                $outputObject = $this->_objectManager->create($schedule->getObjectClass())->load($schedule->getObjectId());
                $message = $outputObject->getData();
            }
            /** Cannot find interfaces nor classes: cannot instantiate object. */            
            else{
                $schedule->setStatus("permanent_error");
                $schedule->setResponse("Cannot instantiat Object");
                $schedule->save();
                continue;
            }
            
            //$this->_logger->info("MESSAGE: ".$message);
            $store_id = $outputObject->getStoreId();
            $envelope=array();            
            $envelope["platform"]=$this->_scopeConfig->getValue(
                'esb/general/platform_code', 
                \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
                $store_id
            );
            $envelope["market"]=$this->_scopeConfig->getValue(
                'esb/general/market_code', 
                \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
                $store_id
            );                        
            $envelope["channeltype"]=$this->_scopeConfig->getValue(
                'esb/general/channeltype', 
                \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
                $store_id
            );                    
            $envelope["message"]=$message;

            $error=false; 

            $outputData = $formatData->format($envelope);
            //$this->_logger->info("PAYLOAD: ".$outputData);

            /** Create HTTP Client. Warning: Zend_HTTP_Client is ZF1 not ZF2.*/
            try{
                if(strlen($schedule->getEndpoint())>0){
                    $client = $this->_httpClientFactory->create();            
                    $client->setUri($schedule->getEndpoint());            
                    $client->setRawData($outputData);            
                    $client->setEncType($formatData->getContentType());                      
                    $client->setMethod(\Zend_Http_Client::POST);
                    $this->_logger->info("Client: ".print_r($client, true));
                    $this->_response = $client->request(); 
                    $this->_logger->info("Response: ".print_r($this->_response, true));
                    $schedule->setResponse($this->_response->getBody());
                    $schedule->setFinishedAt(new \Zend_Db_Expr("NOW()"));
                    if($this->_response->isSuccessful()){
                        $schedule->setStatus("success");                      
                    }else{
                        $this->_logger->error($this->_response->getBody());
                        $schedule->setResponse($this->_response->getBody());
                        $schedule->setStatus("error");
                        $error=true;
                    }                    
                }else{
                    $schedule->setStatus("no_endpoint");
                }
                if(!$error && $schedule->getObjectAfterMethod()!= null && $schedule->getObjectAfterMethod()!= ""){                    
                    $methodName=$schedule->getObjectAfterMethod();
                    $methodArgument=$schedule->getObjectAfterMethodArgument();
                    $this->{$methodName}($outputObject, $methodArgument);
                }  
                $schedule->save();
            }catch(Exception $e){
                $this->_logger->error($e->getMessage());
                $schedule->setStatus("error");
                $schedule->setResponse($e->getMessage());
                $schedule->save();
                $error=true;
            }
            if($error){
                $this->scheduleAgain($schedule);
            }
        }
        return $this;
    }

    /**
     * Validate schedule for processing
     *
     * @param  \FiloBlu\Esb\Model\Schedule $schedule
     * @return array
     */
    protected function validateSchedule($schedule){
        $output=array("result"=>true, "reason" => "");
        switch($schedule->getObjectRepositoryInterface()){
            case "\\Magento\\Sales\Api\\OrderRepositoryInterface":
                $objectRepository=$this->_objectManager->create("\Magento\Sales\Api\OrderRepositoryInterface");
                $order = $objectRepository->get($schedule->getObjectId());
                $skip=false;
                if($order->getStatus()=="closed") $skip=true;
                if($order->getStatus()=="canceled" && $schedule->getObjectAfterMethodArgument()!="canceled") $skip=true;
                if($order->getStatus()=="holded" && $schedule->getObjectAfterMethodArgument()!="holded") $skip=true;
                if($skip){
                    $output["result"]=false;
                    $output["reason"].="Order Status is ".$order->getStatus();
                }
            break;
        }
        return $output;        
    }

    /**
     * Object After Method: update Order Status
     *
     * @param  \Magento\Sales\Api\OrderRepositoryInterface $order
     * @return \FiloBlu\Esb\Model\Schedule $newSchedule
     */
    protected function updateOrderStatus($order, $status){
        $orderManagement=$this->_objectManager->get("\Magento\Sales\Api\OrderManagementInterface");
        $statusHistory=$this->_objectManager->create("\Magento\Sales\Api\Data\OrderStatusHistoryInterface");
        $statusHistory->setStatus($status);
        $comment="ESB Module. ";
        if($this->_response){
            $comment.=$this->parseESBResponse($this->_response->getBody());
        }
        $statusHistory->setComment($comment);       
        $orderManagement->addComment($order->getId(), $statusHistory);
    }

    /**
     * Parse JSON response
     *
     * @param  mixed $message
     * @return string $message
     */
    protected function parseESBResponse($message){
        $decoded=json_decode($message, true);
        if($decoded){
            $output="";
            if(array_key_exists("flow", $decoded)){
                $output.="Flow: ".$decoded["flow"]."<br>";
            }
            if(array_key_exists("message", $decoded)){
                $output.=$decoded["message"].". ";
            }
            if(array_key_exists("description", $decoded)){
                $output.=$decoded["description"].". ";
            }
            if($output){
                return $output;
            }
        }
        return $message;        
    }

    /**
     * Schedule job again. The schedule time depends on the number of previous errors on the same item.
     *
     * @param  \FiloBlu\Esb\Model\Schedule $schedule
     * @return \FiloBlu\Esb\Model\Schedule $newSchedule
     */
    protected function scheduleAgain($schedule){
        $scheduleModel = $this->_scheduleFactory->create();
        $scheduleCollection = $scheduleModel->getCollection()
                        ->addFieldToFilter("status", "error")
                        ->addFieldToFilter("object_class", $schedule->getObjectClass())
                        ->addFieldToFilter("object_repository_interface", $schedule->getObjectRepositoryInterface())
                        ->addFieldToFilter("object_id", $schedule->getObjectId());
        $errors=$scheduleCollection->getSize();
        if($errors < 5){
            $newSchedule = $this->_scheduleFactory->create();
            $minutes = 5 * pow(2 , $errors-1);
            $newSchedule->setData(			
                array(
                    "object_class" => $schedule->getObjectClass(),
                    "object_repository_interface" => $schedule->getObjectRepositoryInterface(),
                    "object_id" => $schedule->getObjectId(),
                    "endpoint" => $schedule->getEndpoint(),
                    "event" => $schedule->getEvent(),
                    "object_after_method" => $schedule->getObjectAfterMethod(),
                    "object_after_method_argument" => $schedule->getObjectAfterMethodArgument(),
                    "status" => "pending",
                    "priority" => $schedule->getPriority(),
                    "executable_after" => new \Zend_Db_Expr('TIMESTAMPADD(MINUTE,'.$minutes.', NOW())')
			    )
            );
            $newSchedule->save();
            return $newSchedule;
        }else{
            $schedule->setStatus("permanent_error");
            $schedule->save();
            $this->_logger->error("Won't schedule again. More than 4 errors while scheduling. ".json_encode($schedule));
            return false;
        }
    }
    
    /**
     * If gift wrap is selected, add linked product to order, see config esb/gwmapping/gwproducts 
     *
     * @param  \Magento\Sales\Api\OrderRepositoryInterface $order
     */
    protected function addGiftWrapProduct($order){
        
        $products=unserialize($this->_scopeConfig->getValue(
            'esb/gwmapping/gwproducts', 
            \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
            $order->getStoreId()
        ));  
        $productSku = null;
        foreach($products as $productData): error_log(print_r($productData,true));
            if($productData['wrapping_id'] == $order->getGwId() || $productData['wrapping_id'] == "*"):
                $productSku = $productData['product_sku'];
                break;
            endif;
        endforeach;
        
            try{
                 
                $product = $this->_productFactory->create()->loadByAttribute('sku', $productSku);
                $orderItem = $this->_orderItemFactory->create();

                $orderItem->setProductId(236);
                $orderItem->setSku($product->getSku());
                $orderItem->setName($product->getName());
                $orderItem->setQtyOrdered(1);
                $orderItem->setQtyInvoiced(1);
                $orderItem->setProductType($product->getTypeId());
                $orderItem->setBasePrice($order->getGwBasePrice());
                $orderItem->setPrice($order->getGwPrice());
                $orderItem->setBasePriceInclTax($order->getGwBasePriceInclTax());
                $orderItem->setPriceInclTax($order->getGwPriceInclTax());
                $orderItem->setBaseTaxAmount($order->getGwBaseTaxAmount());
                $orderItem->setTaxAmount($order->getGwTaxAmount());
                $orderItem->setTaxInvoiced($order->getGwTaxAmount());
                $orderItem->setBaseRowInvoiced($order->getGwBasePrice());
                $orderItem->setBaseRowTotal($order->getGwBasePrice());
                $orderItem->setRowTotal($order->getGwBasePrice());
                $orderItem->setBaseRowTotalInclTax($order->getGwBasePriceInclTax());
                $orderItem->setTaxPercent(round($order->getGwTaxAmount()/$order->getGwBasePrice()*100));
                

                $orderItem->setBaseAmountRefunded("0.0000");
                $orderItem->setBaseDiscountAmount("0.0000");
                $orderItem->setBaseDiscountTaxCompensationAmount("0.0000");
                $orderItem->setBaseDiscountTaxCompensationInvoiced("0.0000");
                $orderItem->setDiscountAmount("0.0000");
                $orderItem->setDiscountInvoiced("0.0000");
                $orderItem->setDiscountPercent("0.0000");
                
                //$extensionAttributes = $this->_objectManager->create("\Magento\Sales\Api\Data\OrderItemExtensionFactory");
                //$extensionAttributes->setData("kpi_full_price", $order->getGwBasePriceInclTax());
                //$orderItem->setExtensionAttributes($extensionAttributes);

                $order->addItem($orderItem);    
                       
            } catch(\Exception $e) {
                error_log($e->getMessage());
            }
                                   
    }
    
}